# %% Imports
from ngsolve import *
from ngsolve import CoefficientFunction as Coeff
from netgen.csg import *
import numpy as np

# %% Energy Function
w = []
b = []
w.append(Coeff(tuple(np.zeros((6,32)).flatten()), dims=(6,32)))
b.append(Coeff(tuple(np.zeros((1,32)).flatten()), dims=(1,32)))
w.append(Coeff(tuple(np.ones((32,32)).flatten()), dims=(32,32)))
b.append(Coeff(tuple(np.zeros((1,32)).flatten()), dims=(1,32)))
w.append(Coeff(tuple(np.zeros((32,1)).flatten()), dims=(32,1)))
b.append(Coeff(tuple(np.zeros((1,1)).flatten()), dims=(1,1)))

def green_lag (disp):
    I = Id(disp.dim)
    F = I + Grad(disp)
    C = F.trans * F     
    E = 0.5*(C-I)
    return E

def Potential (E):
    E_star = list()
    E_star.append(E[0,0])
    E_star.append(E[1,1])
    E_star.append(E[2,2])
    E_star.append(0.5 * (E[2,1]+E[1,2]))
    E_star.append(0.5 * (E[2,0]+E[0,2]))
    E_star.append(0.5 * (E[0,1]+E[1,0]))
    W = Coeff(tuple(E_star), dims=(1,6))    
    for w_s, b_s in zip(w, b):
        preFunc = W * w_s + b_s
        ones = Coeff(tuple(np.ones(preFunc.dim)), dims=tuple(preFunc.dims))
        W = preFunc / (ones+exp(-1*preFunc))      
    return W

def PK_stress (disp):
    I = Id(disp.dim)
    F = I + Grad(disp)
    C = F.trans * F     
    E = 0.5*(C-I)
    W = Potential(E)
    P = W.Diff(F)
    return P

# %% Geometry
left  = Plane (Pnt(0,0,0), Vec(-1,0,0)).bc('left')
right = Plane (Pnt(1,1,1), Vec( 1,0,0)).bc('right')
front = Plane (Pnt(0,0,0), Vec(0,-1,0)).bc('bot')
back  = Plane (Pnt(1,1,1), Vec(0, 1,0)).bc('top')
bot   = Plane (Pnt(0,0,0), Vec(0,0,-1)).bc('front')
top   = Plane (Pnt(1,1,1), Vec(0,0, 1)).bc('back')
cube = left * right * front * back * bot * top
geo = CSGeometry()
geo.Add (cube)
mesh = Mesh(geo.GenerateMesh(maxh=0.2))

eval_point = mesh(.5, .5, .5)

# %% Equation
fes = VectorH1(mesh, order=1, dirichletx='left|right', dirichlety='bot', dirichletz='back')
u  = fes.TrialFunction()
a = BilinearForm(fes, symmetric=True)
a += Variation(Potential(green_lag(u)).Compile()*dx)
gfu = GridFunction(fes)


# %% Solving
Draw (gfu, mesh, "disp")
SetVisualization (deformation=True) 

W_sim = []
P_sim = []

for loadstep in np.linspace(0, -0.2, 50):    
    print ("\nDeform:", loadstep)  
    gfu.components[0].Set(loadstep, definedon=mesh.Boundaries('right'))
    conv, steps = solvers.Newton(a, gfu, fes.FreeDofs(), inverse='', maxit=25) 
    if conv:
        break      
    P = PK_stress(gfu)
    W = Potential(green_lag(gfu))
    
    Draw (gfu, mesh, "disp")
    Draw (W, mesh, "EnDens")
    SetVisualization (deformation=True)  
    
    P_sim.append(P(eval_point))
    W_sim.append(W(eval_point))